// ADO DataSet component v0.9
// Written by Larry Nezar (LarryN@nda.agric.za)
// Date of last modification: 28/7/1999

unit ADODataSet;

interface

uses
  Windows, SysUtils, Classes, Db, ADODB_TLB, OleCtrls;

type TadoInfo = record
  BookMarkInt: OleVariant;
  BookmarkFlag: TBookmarkFlag;
end;
PadoInfo = ^TadoInfo;
PMyStamp = ^TTimeStamp;

ADOError = class(Exception);
TMsgEvent = procedure(Sender: TObject; msg: String) of object;
TrsStatus = (myNormal,myEOF,myBOF);

type
  TADODataSet = class(TDataSet)
  private
    { Private declarations }
    fFiltered: Boolean;
    fConnectParams: String;
    fRecordSetSQL: String;
    fProvider: String;
    fSource: String;
    fCatalog : String;
    fUsername: String;
    fPassword: String;
    IsOpen: Boolean;
    EMode: Boolean;
    NewRec: OleVariant;
    StartInfo, StartCalc: Integer;
    SavedBookmark: OleVariant;
    SaveStatus: TrsStatus;
    ADOConnection: _Connection;
    ADORecordset: _Recordset;
    fmessage: TMsgEvent;
    function ADOFieldType(ID: DataTypeEnum) :TFieldType;
    function GetActiveRecordBuffer: PChar;
    procedure SaveADOStatus;
    procedure RestoreADOStatus;
  protected
    { Protected declarations }
    procedure InternalOpen; override;
    procedure InternalClose; override;
    function IsCursorOpen: Boolean; override;
    function GetCanModify: Boolean; override;
    function GetRecordSize: Word; override;
    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure InternalFirst; override;
    procedure InternalLast; override;
    procedure InternalInitFieldDefs; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
    function GetBlobFieldData(FieldNo: Integer; var Buffer: TBlobByteData): Integer; override;
    function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
    function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    function GetBookMarkFlag(Buffer: PChar): TBookMarkFlag; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure ClearCalcFields(Buffer: PChar);
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalDelete; override;
    procedure InternalPost; override;
    procedure SetFiltered(Value: Boolean); override;
    procedure DoMessage(Msg: string); dynamic;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    { Published declarations .. redeclare all}
    property Active;
    property Filtered: Boolean read fFiltered write SetFiltered default false;
    property Filter;
    property ConnectParams: String read fConnectParams write fConnectParams;
    property Provider: String read fProvider write fProvider;
    property Source: String read fSource write fSource;
    property Catalog: String read fCatalog write fCatalog;
    property Username: String read fUsername write fUsername;
    property Password: String read fPassword write fPassword ;
    property RecordSetSQL: String read fRecordSetSQL write fRecordSetSQL;
    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;
    property OnMessage: TMsgEvent read fmessage write fmessage;
  end;

type
  TadoBlobStream = class(TStream)
  private
    FField: TBlobField;
    FDataSet: TADODataSet;
    FMode: TBlobStreamMode;
    FFieldNo: Integer;
    FOpened: Boolean;
    FModified: Boolean;
    FPosition: Longint;
    FBlobData: TBlobData;
    function GetBlobSize: Longint;
  public
    constructor Create(Field: TBlobField; Mode: TBlobStreamMode);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
//    procedure Truncate;
  end;

procedure Register;

implementation

constructor TADODataSet.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
end;

destructor TADODataSet.Destroy;
begin
  inherited Destroy;
end;

procedure TADODataSet.InternalOpen;
begin
  // CacheBlobs:=False;   // ADO does the cacheing
  if not Assigned(ADOConnection) then
  begin
    ADOConnection := CoConnection.Create;
  end;
  ADOConnection.CursorLocation := adUseClient;
//  ADOConnection.Provider := fProvider;
  if fCatalog='' then
    ADOConnection.Open('Provider = ' + fProvider + ';' +
                     'Data Source = ' + fSource + ';' +
                     'User ID = ' + fUsername + ';' +
                     'Password = ' + fPassword + ';' +
                     fConnectParams, '', '',-1)
  else
    ADOConnection.Open('Provider = ' + fProvider + ';' +
                     'Data Source = ' + fSource + ';' +
                     'Initial Catalog = ' + fCatalog + ';' +
                     'User ID = ' + fUsername + ';' +
                     'Password = ' + fPassword + ';' +
                     fConnectParams, '', '',-1);
  InternalInitFieldDefs;
  if DefaultFields then CreateFields;
  BindFields(True);
  StartInfo := 0;               // No actual records
  StartCalc := StartInfo + SizeOf(TadoInfo);
  IsOpen := True;
  EMode:=True;
  DoMessage('Blobs='+IntToStr(BlobFieldCount))
end;

function TADODataSet.ADOFieldType(ID: DataTypeEnum) :TFieldType;
begin
//  DoMessage('DataType: '+IntToStr(ID));
  case ID of
    adInteger:
      ADOFieldType := ftInteger;
    adVarChar, adChar:
      ADOFieldType := ftString;
    adLongVarChar:
      ADOFieldType := ftMemo;
    adDBTimeStamp:
      ADOFieldType := ftDateTime;
    adDouble, adSingle, adNumeric:
      ADOFieldType := ftFloat;
    adLongVarBinary:
      ADOFieldType := ftBlob;
    adCurrency:
      ADOFieldType := ftCurrency;
    adSmallInt, adTinyInt:
      ADOFieldType := ftSmallInt;
    adVarBinary:
      ADOFieldType := ftVarBytes;
    else
      ADOFieldType := ftUnknown; // wrong
  end;
end;

procedure TADODataSet.InternalInitFieldDefs;
var
  RowsAffected : OleVariant;
  nFields, I : Integer;
begin
  if not Assigned(ADORecordset) then
  begin
    ADORecordSet := CoRecordSet.Create;
    ADORecordSet.CursorLocation := adUseClient;
    ADORecordSet.Open(fRecordSetSQL, ADOConnection, adOpenStatic, adLockOptimistic,-1)
  end;
  InternalFirst;    // make sure BOF is true;
  FieldDefs.Clear;
  nFields := ADORecordSet.Fields.Count;
  for I := 0 to nFields-1 do
  begin
    If ADORecordSet.Fields[I].Type_ in [adVarChar, adVarBinary, adChar] then
      FieldDefs.Add(ADORecordSet.Fields[I].Name,
                    ADOFieldType(ADORecordSet.Fields[I].Type_),
                    ADORecordSet.Fields[I].DefinedSize, False)
    else
      FieldDefs.Add(ADORecordSet.Fields[I].Name,
                    ADOFieldType(ADORecordSet.Fields[I].Type_),
                    0, False);
  end;
end;

procedure TADODataSet.InternalClose;
begin
  IsOpen := False;
  BindFields(False);
  if DefaultFields then DestroyFields;
  ADORecordSet.Close;
  ADORecordSet := nil;
  ADOConnection.Close;
  ADOConnection := nil;
end;

function TADODataSet.IsCursorOpen: Boolean;
begin
  Result := IsOpen;
end;

function TADODataSet.GetCanModify: Boolean;
begin
  Result := true;           // beware here...
end;

function TADODataSet.GetRecordSize: Word;
begin
  Result := SizeOf(TadoInfo) + CalcFieldsSize;
end;

function TADODataSet.AllocRecordBuffer: PChar;
begin
  GetMem(Result, GetRecordSize);
end;

procedure TADODataSet.FreeRecordBuffer( var Buffer: PChar);
begin
  FreeMem(Buffer, GetRecordSize);
end;

procedure TADODataSet.InternalFirst;
begin
  DoMessage('InternalFirst');
  ADORecordSet.MoveFirst;
  ADORecordSet.MovePrevious;    // make sure bof is true
end;

procedure TADODataSet.InternalLast;
begin
  DoMessage('InternalLast');
    ADORecordSet.MoveLast;
  if EMode then ADORecordSet.MoveNext;        // make sure eof is true
  EMode:=True;//  if ADORecordset.EditMode = adEditAdd then ADORecordset.MoveLast;
end;

function TADODataSet.GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
begin
  DoMessage('GetRecord');
  Result := grOK;
  case GetMode of
    gmCurrent: begin
                 DoMessage('gmCurrent');
                 if ADORecordSet.EOF or ADORecordSet.BOF then Result:=grError;
               //  if (ADORecordSet.Status and adRecDeleted)= adRecDeleted then Result:=grError;
               end;
    gmNext:  begin
//               If not EMode then InternalDelete;
//               EMode:=True;
//               DoMessage('gmNext');
               if not ADORecordSet.EOF then begin
                 if ADORecordSet.BOF then begin
                   ADORecordSet.MoveFirst;
                   EMode:=true;
                 end else begin
                   ADORecordSet.MoveNext;
                   EMode:=true;
                 end;
                 if ADORecordSet.EOF then begin
                   Result:=grEOF;
                 end;
               end else begin
                 Result:=grEOF;
               end;
             end;
    gmPrior: begin
//  //             If not EMode then InternalDelete;
//  //            EMode:=True;
//               DoMessage('gmPrior');
               if not ADORecordSet.BOF then begin
                 if ADORecordSet.EOF then begin
                   ADORecordSet.MoveLast;
                   EMode:=true;
                 end else begin
                   ADORecordSet.MovePrevious;
                   EMode:=true;
                 end;
                 if ADORecordSet.BOF then begin
                   Result:=grBOF;
                 end;
               end else begin
                 Result:=grBOF;
               end;
             end;
  end;
  if Result = grOK then
  begin
//    DoMessage('SetToRecord:'+IntToStr(ADORecordset.fields[0].Value));
    with PadoInfo(Buffer+StartInfo)^ do
    begin
      BookmarkFlag := bfCurrent;
      BookmarkInt := ADORecordSet.Bookmark;
    end;
    ClearCalcFields(Buffer);
    GetCalcFields(Buffer);
  end;
end;

procedure TADODataSet.ClearCalcFields(Buffer: PChar);
begin
  FillChar(Buffer[StartCalc], CalcFieldsSize, 0);
end;

function TADODataSet.GetBookMarkFlag(Buffer: PChar): TBookmarkFlag;
begin
  DoMessage('GetBookMarkFlag');
  Result := PadoInfo(Buffer+StartInfo).BookmarkFlag;
end;

procedure TADODataSet.InternalSetToRecord(Buffer: PChar);
begin
  DoMessage('InternalSetToRecord');
  ADORecordSet.Bookmark := PadoInfo(Buffer+StartInfo).BookmarkInt;
end;

function TADODataSet.GetActiveRecordBuffer: PChar;
begin
  DoMessage('GetActiveRecordBuffer');
  case State of
    dsBrowse:
      if IsEmpty then Result := nil else Result := ActiveBuffer;
    dsCalcFields:
      Result := CalcBuffer;
//.   dsFilter:
//.     Result := FFilterBuffer;
    dsEdit, dsInsert:
      Result := ActiveBuffer;
  else
    Result := nil;
  end;
end;

procedure TADODataSet.SetFieldData(Field: TField; Buffer: Pointer);
var
  DestBuffer: PChar;
  tmpString: string;
  DTmp: TDateTime;
begin
  DestBuffer := GetActiveRecordBuffer;
  if (Field.FieldKind = fkCalculated) or (Field.FieldKind = fkLookup) then
  begin
    Inc(DestBuffer, StartCalc+Field.Offset);
    Boolean(DestBuffer[0]):=(Buffer<>nil);
    if Boolean(DestBuffer[0]) then
      CopyMemory(@DestBuffer[1], Buffer, Field.DataSize);
  end
  else begin
    EMode:=True;
    SaveADOStatus;
    ADORecordSet.Bookmark:=PadoInfo(DestBuffer+StartInfo).BookmarkInt;
    case Field.DataType of
      ftInteger, ftSmallInt:
        if Buffer<>nil then begin
          ADORecordSet.Fields.Item[Field.FieldName].Value := Integer(Buffer^);
          // ADORecordSet.Update(Field.FieldName,Integer(Buffer^));
        end;
      ftFloat, ftCurrency:
        if Buffer<>nil then begin
          ADORecordSet.Fields.Item[Field.FieldName].Value := Double(Buffer^);
        end;
      ftDateTime:
        if Buffer<>nil then begin
          // TwoDigitYearCenturyWindow:=50;
          // ShortDateFormat:='mm/dd/yyyy';
          DTmp:=TimeStampToDateTime(MSecsToTimeStamp(Double(Buffer^)));
          ADORecordSet.Fields.Item[Field.FieldName].Value := DateTimeToStr(DTmp);
        end;
      ftString:
        if Buffer<>nil then begin
          tmpString:=PChar(Buffer);
          // ADORecordSet.Update(Field.FieldName, tmpString);
          ADORecordSet.Fields.Item[Field.FieldName].Value := tmpString;
          // StrCopy(PChar(string(ADORecordSet.Fields.Item[Field.FieldName].Value)),PChar(Buffer));
        end;
    end;
    RestoreADOStatus;
  end;
end;

function TADODataSet.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
  SourceBuffer: PChar;
  DTmp: TDateTime;
  FTmp: Double;
  V: Variant;
  tmpBytes: Pointer;
begin
  DoMessage('GetFieldData :'+Field.FieldName);
  Result:=False;
  SourceBuffer:=GetActiveRecordBuffer;   // called here
  if not IsOpen or (SourceBuffer = nil) then Exit;
  if (Field.FieldKind = fkCalculated) or (Field.FieldKind = fkLookup) then
  begin
    Inc(SourceBuffer, StartCalc+Field.Offset);
    if (SourceBuffer[0]=#0) or (Buffer=nil) then Exit
    else CopyMemory(Buffer, @SourceBuffer[1], Field.DataSize);
  end
  else begin  // if EMode then
    SaveADOStatus;
   // try
    ADORecordSet.Bookmark := PadoInfo(SourceBuffer+StartInfo).BookmarkInt;
    case Field.DataType of
      ftInteger, ftSmallInt:
        if ADORecordSet.Fields.Item[Field.FieldName].Value <> null then
          Integer(Buffer^):= Integer(ADORecordSet.Fields.Item[Field.FieldName].Value)
        else
          Integer(Buffer^):=0;
      ftFloat,ftCurrency:
       begin
//        DoMessage(string(ADORecordSet.Fields.Item[Field.FieldName].Value));
        if ADORecordSet.Fields.Item[Field.FieldName].Value <> null then
        begin
//        try      // somehow test for null does not work with adNumeric types
          Double(Buffer^):= double(ADORecordSet.Fields.Item[Field.FieldName].Value);
        end else begin
          //        except
          Double(Buffer^):=0.0;
        end;
       end;
      ftVarBytes: begin
        { if ADORecordSet.Fields.Item[Field.FieldName].Value <> null then begin
     //     DoMessage('VarBinType = '+IntToStr(VarType(ADORecordSet.Fields.Item[Field.FieldName].Value)));
          V:=ADORecordSet.Fields.Item[Field.FieldName].Value;
          tmpBytes := VarArrayLock(v);
          move(PChar(tmpBytes), Buffer, ADORecordSet.Fields.Item[Field.FieldName].ActualSize);   // Size is wrong to use here
          VarArrayUnlock(v);
        end else
          ;  }
       end;
      ftString: begin
//        DoMessage('ftString');
        if ADORecordSet.Fields.Item[Field.FieldName].Value <> null then
          StrCopy(PChar(Buffer), PChar(string(ADORecordSet.Fields.Item[Field.FieldName].Value)))
        else
          StrCopy(PChar(Buffer), '');
       end;
      ftDateTime: begin
//        DoMessage('ftDateTime:'+IntToStr(Field.DataSize));
        TwoDigitYearCenturyWindow:=50;
        ShortDateFormat:='mm/dd/yyyy';
//        DoMessage(string(ADORecordSet.Fields.Item[Field.FieldName].Value));
        // DoMessage(DateTimeToStr(Tdatetime(ADORecordSet.Fields.Item[Field.FieldName].Value)));
        if ADORecordSet.Fields.Item[Field.FieldName].Value <> null then
        begin
          DTmp := StrToDateTime(string(ADORecordSet.Fields.Item[Field.FieldName].Value));
          Double(Buffer^):=TimeStampToMSecs(DateTimeToTimeStamp(DTmp));
        end
        else
          ;
        // DoMessage(DateTimeToStr(TimeStampToDateTime(TTimeStamp(Buffer^))));
       end;
      ftMemo: begin
//        DoMessage('ftMemo');
        if Buffer <> nil then begin
          // if ADORecordSet.Fields.Item[Field.FieldName].Value <> null then begin
          // DoMessage(string(ADORecordSet.Fields.Item[Field.FieldName].Value));
          //  StrCopy(PChar(Buffer), PChar(string(ADORecordSet.Fields.Item[Field.FieldName].Value)));
          //  DoMessage(string(Buffer));
          // end else
          //  StrCopy(PChar(Buffer), '');
        end else begin
          // Buffer:=PChar(string(ADORecordSet.Fields.Item[Field.FieldName].Value));
          // DoMessage(string(ADORecordSet.Fields.Item[Field.FieldName].Value));
        end;
       end;
      ftBlob: begin
//        DoMessage('ftGraphic');
        if Buffer <> nil then begin
          // if ADORecordSet.Fields.Item[Field.FieldName].Value <> null then begin
          // DoMessage(string(ADORecordSet.Fields.Item[Field.FieldName].Value));
          //  StrCopy(PChar(Buffer), PChar(string(ADORecordSet.Fields.Item[Field.FieldName].Value)));
          //  DoMessage(string(Buffer));
          // end else
          //  StrCopy(PChar(Buffer), '');
        end else begin
          // Buffer:=PChar(string(ADORecordSet.Fields.Item[Field.FieldName].Value));
          // DoMessage(string(ADORecordSet.Fields.Item[Field.FieldName].Value));
        end;
       end;
    end;
   // finally
    RestoreADOStatus;
   // end;
  end;
  Result:=true;
end;

procedure TADODataSet.InternalPost;
begin
  DoMessage('Post updates...');
  ADORecordSet.Bookmark:=PadoInfo(ActiveBuffer+StartInfo).BookMarkInt;
  ADORecordSet.Update(EmptyParam, EmptyParam);
  EMode:=True;
end;

procedure TADODataSet.InternalDelete;
begin
  DoMessage('Delete record...');
  ADORecordSet.Delete(adAffectCurrent);
  IF ADORecordSet.EOF then ADORecordSet.MoveLast
    else ADORecordSet.MoveNext;
  EMode:=True;
end;

procedure TADODataSet.InternalAddRecord(Buffer: Pointer; Append: Boolean);
begin
  DoMessage('Add new record...');
//  ADORecordSet.Bookmark:=PadoInfo(PChar(Buffer)+StartInfo).BookMarkInt;
//  ADORecordSet.AddNew(EmptyParam, EmptyParam);
//  PadoInfo(Buffer+StartInfo).BookMarkInt:=ADORecordSet.Bookmark;
  EMode:=True;
end;

procedure TADODataSet.InternalInitRecord(Buffer: PChar);
begin
  DoMessage('InternalInitRecord');
  FillChar(Buffer^, GetRecordSize, 0);
  ADORecordSet.AddNew(0, null);
  ADORecordSet.Update(EmptyParam, EmptyParam);
  PadoInfo(Buffer+StartInfo).BookmarkInt:=ADORecordSet.Bookmark;
//  NewRec:=PadoInfo(Buffer+StartInfo).BookmarkInt;
  EMode:=False;
end;

procedure TADODataSet.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
  DoMessage('SetBookmarkFlag');
  PadoInfo(Buffer+StartInfo).BookmarkFlag:= Value;
end;

procedure TADODataSet.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
// This I can't get right
  DoMessage('GetBookmarkData');
//  SetLength(String(Data), SizeOf(OleVariant));
// //  Move(PadoInfo(Buffer).BookmarkInt,Data,16);
//  POleVariant(Data)^ := PadoInfo(Buffer+StartInfo)^.BookmarkInt;
end;

procedure TADODataSet.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
// neither can I get this right
  DoMessage('SetBookmarkData');
// //  Move(Data,PadoInfo(Buffer).BookmarkInt,16);
//  PadoInfo(Buffer+StartInfo)^.BookmarkInt:=POleVariant(Data)^;
end;

procedure Register;
begin
  RegisterComponents('Samples', [TADODataSet]);
end;

procedure TADODataSet.DoMessage(Msg: string);
begin
  if Assigned(fmessage) then
    fmessage(Self, Msg);
end;

procedure TADODataSet.InternalGotoBookmark(Bookmark: Pointer);
begin
  DoMessage('Internalgotobookmark');
  ADORecordSet.Bookmark := PadoInfo(Bookmark)^.BookmarkInt;
end;

function TADODataSet.GetBlobFieldData(FieldNo: Integer; var Buffer: TBlobByteData): Integer;
var
  SourceBuffer: PChar;
  SavedBookmark: OleVariant;
  SaveStatus: TrsStatus;
begin
//  DoMessage('BlobFetchData');
  SourceBuffer:=GetActiveRecordBuffer;   // called here
    SaveADOStatus;
    ADORecordSet.Bookmark := PadoInfo(SourceBuffer+StartInfo).BookmarkInt;
    // PChar(string(ADORecordSet.Fields[FieldNo].Value))
    StrCopy(PChar(Buffer), 'Hello');
    Result:=ADORecordSet.Fields[Fieldno-1].ActualSize;
  RestoreADOStatus;
end;

function TADODataSet.CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream;
var
 a: PChar;
begin
//  DoMessage('CreateBlobStream');
  Result:=TadoBlobStream.Create(Field as TBlobField, Mode);
end;

{---------------- TadoBlobStream ------------------}

constructor TadoBlobStream.Create(Field: TBlobField; Mode: TBlobStreamMode);
begin
  FMode := Mode;
  FField := Field;
  FDataSet := FField.DataSet as TADODataSet;
  FFieldNo := FField.FieldNo;
  FOpened := True;
end;

destructor TadoBlobStream.Destroy;
begin
  if FOpened then
  begin
//    FDataSet.DoMessage('Destroy Stream');
    if FModified then FField.Modified := True;
  end;
end;

function TadoBlobStream.GetBlobSize: Longint;
var
 SourceBuffer: PChar;
begin
  Result := 0;
  SourceBuffer:=FDataSet.GetActiveRecordBuffer;   // called here
  FDataSet.SaveADOStatus;
  FDataSet.ADORecordSet.Bookmark := PadoInfo(SourceBuffer+FDataSet.StartInfo).BookmarkInt;

//  FDataSet.DoMessage('GetBlobSize');
  if FOpened then
    Result:=FDataSet.ADORecordSet.Fields[FFieldNo-1].ActualSize;
  FDataSet.RestoreADOStatus;
//  FDataSet.DoMessage('GetBlobSize='+IntToStr(Result));
end;

function TadoBlobStream.Read(var Buffer; Count: Integer): Longint;
var
 tmpString: Pstring;
 SourceBuffer: PChar;
 tmpBytes: Pointer;
 V: Variant;
begin
  Result := 0;
//  FDataSet.DoMessage('Blob.Read');
  if FOpened then
  begin
    SourceBuffer:=FDataSet.GetActiveRecordBuffer;   // called here
    FDataSet.SaveADOStatus;
    FDataSet.ADORecordSet.Bookmark := PadoInfo(SourceBuffer+FDataSet.StartInfo).BookmarkInt;

//    FDataSet.DoMessage('Read.Count: '+IntToStr(Count));
    case FField.DataType of
      ftMemo:
        begin
          tmpString := Pstring(ansistring(FDataSet.ADORecordSet.Fields[FFieldNo-1].Value));
          move(PChar(tmpString)[FPosition], Buffer, Count);     // move(tmpString^, Buffer, Count);
          inc(FPosition, Count);
          Result:=Count;
        end;
      ftBlob, ftGraphic:
        begin
          V:=FDataSet.ADORecordSet.Fields[FFieldNo-1].Value;
          tmpBytes := VarArrayLock(v);
          move(PChar(tmpBytes)[FPosition], Buffer, Count);
          inc(FPosition, Count);
          VarArrayUnlock(v);
          Result:=Count;
        end;
      end;
    FDataSet.RestoreADOStatus;
  end;
end;

function TadoBlobStream.Seek(Offset: Integer; Origin: Word): Longint;
begin
//  FDataSet.DoMessage('Seek Blob offset: '+IntToStr(Offset)+':'+IntToStr(Origin));
  case Origin of
    0: FPosition := Offset;
    1: Inc(FPosition, Offset);
    2: FPosition := GetBlobSize + Offset;
  end;
  Result := FPosition;
//  FDataSet.DoMessage('FPos = '+IntToStr(Result));
end;
{
procedure TadoBlobStream.Truncate;
begin
  // ???
end;
}
function TadoBlobStream.Write(const Buffer; Count: Integer): Longint;
var
 tmpString: Pstring;
 DestBuffer: PChar;
 tmpBytes: Pointer;
 tmpChar: ansistring;
 emptyString: string;
 V: Variant;
begin
  Result := 0;
//  FDataSet.DoMessage('Blob.Write');
  if FOpened then
  begin
    emptystring:='';
    DestBuffer:=FDataSet.GetActiveRecordBuffer;   // called here
    FDataSet.SaveADOStatus;
    FDataSet.ADORecordSet.Bookmark := PadoInfo(DestBuffer+FDataSet.StartInfo).BookmarkInt;

//    FDataSet.DoMessage('Write.Count: '+IntToStr(Count));
    case FField.DataType of
      ftMemo:
        begin
          if FPosition<>0 then begin
            tmpString := Pstring(ansistring(FDataSet.ADORecordSet.Fields[FFieldNo-1].Value));
            SetLength(string(tmpString), Length(string(tmpString))+Count);
          end else begin
            SetLength(emptyString, Count);
            tmpString:=Pstring(emptyString);
          end;
          move(Buffer, PChar(tmpString)[FPosition], Count);     
          inc(FPosition, Count);
          tmpChar:=PChar(tmpString);
          FDataSet.ADORecordSet.Fields[FFieldNo-1].Value:=tmpChar;
          Result:=Count;
        end;
      ftBlob, ftGraphic:
        begin
          if FPosition<>0 then
            v:=FDataSet.ADORecordSet.Fields[FFieldNo-1].Value
          else
            v:=varArrayCreate([0,0], varByte);
          VarArrayRedim(v, VarArrayHighBound(v, 1)+Count);
          tmpBytes := VarArrayLock(v);
          move(Buffer, PChar(tmpBytes)[FPosition], Count);
          inc(FPosition, Count);
          VarArrayUnlock(v);
          FDataSet.ADORecordSet.Fields[FFieldNo-1].Value:=v;
          Result:=Count;
        end;
      end;
    FDataSet.RestoreADOStatus;
  end;
end;

procedure TADODataSet.SaveADOStatus;  // save current position
begin
  if ADORecordSet.EOF then SaveStatus:=myEOF
    else if ADORecordSet.BOF then SaveStatus:=myBOF
      else SaveStatus:=myNormal;
  if SaveStatus=myNormal then SavedBookmark := ADORecordSet.Bookmark;
end;

procedure TADODataSet.RestoreADOStatus;     // restore current position
begin
  case SaveStatus of
    myNormal: ADORecordSet.Bookmark := SavedBookmark;
    myEOF: InternalLast;
    myBOF: InternalFirst;
  end;
end;


procedure TADODataSet.SetFiltered(Value: Boolean);
begin
  If Value=true then begin
    If (Filter<>'')and (Active) then begin
      ADORecordSet.Filter := Filter;
      fFiltered:=Value;
    end;
  end
  else begin
    ADORecordSet.Filter := adFilterNone;
    fFiltered:=Value;
  end;
  Refresh;
end;

end.
